/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <GL/gl.h>
//#include <GL/glu.h>

#define TEXPREF_NONE			0x0000
#define TEXPREF_MIPMAP			0x0001	// generate mipmaps
#define TEXPREF_LINEAR			0x0002	// force linear
#define TEXPREF_NEAREST			0x0004	// force nearest
#define TEXPREF_ALPHA			0x0008	// allow alpha
#define TEXPREF_PAD				0x0010	// allow padding (UNUSED)
#define TEXPREF_PERSIST			0x0020	// never free (UNUSED)
#define TEXPREF_OVERWRITE		0x0040	// overwrite existing same-name texture
#define TEXPREF_NOPICMIP		0x0080	// always load full-sized
#define TEXPREF_FULLBRIGHT		0x0100	// use fullbright mask palette
#define TEXPREF_NOBRIGHT		0x0200	// use nobright mask palette
#define TEXPREF_CONCHARS		0x0400	// use conchars palette
#define TEXPREF_WARPIMAGE		0x0800	// resize this texture when warpimagesize changes (UNUSED)

enum srcformat {SRC_INDEXED, SRC_LIGHTMAP, SRC_RGBA};

typedef struct gltexture_s {
//managed by texture manager
	int		texnum;
	struct gltexture_s	*next;
	model_t				*owner;
//managed by image loading
	char				name[64];
	int					width; //size of image as it exists in opengl
	int					height; //size of image as it exists in opengl
	unsigned			flags;
	enum srcformat		source_format; //format of pixel data (indexed, lightmap, or rgba)
	int					source_width; //size of image in source data
	int					source_height; //size of image in source data
	unsigned short		source_crc; //generated by source data before modifications
} gltexture_t;

extern	model_t	*loadmodel;

// vid_*gl*.c
void GL_BeginRendering (int *x, int *y, int *width, int *height);
void GL_EndRendering (void);

// gl_rmain.c
qboolean R_CullBox (vec3_t mins, vec3_t maxs);
qboolean R_CullModelForEntity (entity_t *e);

// gl_draw.c
void GL_Upload8 (gltexture_t *glt, byte *data);
void GL_Upload32 (gltexture_t *glt, unsigned *data);
void GL_UploadLightmap (gltexture_t *glt, byte *data);
void GL_FreeTexture (gltexture_t *free);
void GL_FreeTextures (model_t *owner);
gltexture_t *GL_LoadTexture (model_t *owner, char *name, int width, int height, enum srcformat format, byte *data, unsigned flags);
gltexture_t *GL_FindTexture (model_t *owner, char *name);
void GL_Set2D (void);
void Draw_Pic (int x, int y, qpic_t *pic);
void Draw_SubPic (int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height);
void GL_SelectTexture (GLenum target);
void GL_Bind (gltexture_t *texture);
void GL_DisableMultitexture (void);
void GL_EnableMultitexture (void);
void GL_Init (void);
void GL_SetPalette (byte *palette);

// gl_mesh.c
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr);

// gl_rmisc
void R_InitTranslatePlayerTextures (void);
void R_TranslatePlayerSkin (int playernum);

// gl_part.c
void R_InitParticles (void);
void R_DrawParticles (void);
void R_ClearParticles (void);

// gl_refrag.c
void R_StoreEfrags (efrag_t **ppefrag);
 
// gl_rlight.c
void R_AnimateLight (void);
int R_LightPoint (vec3_t p);
void R_RenderDlights (void);
void R_MarkLights (dlight_t *light, int bit, mnode_t *node); 

// gl_rsurf.c
void R_CullSurfaces (void);
void R_DrawBrushModel (entity_t *e, qboolean water);
void R_DrawWorld (void);
void R_DrawTextureChains_Water (void);
void R_DrawSequentialPoly (msurface_t *s);
void R_DrawSequentialWaterPoly (msurface_t *s);
void DrawGLPoly34 (glpoly_t *p);
void DrawGLPoly56 (glpoly_t *p);
void GL_BuildLightmaps (void);

// gl_warp.c
void R_UpdateWarpTextures (void);

void R_InitSkyBox (void);
void R_SkyNewMap (void);
void R_DrawSky (void);
void R_LoadSkyBox (char *skybox);

void Fog_ParseServerMessage (void);
float *Fog_GetColor (void);
float Fog_GetDensity (void);
void Fog_EnableGFog (void);
void Fog_DisableGFog (void);
void Fog_StartAdditive (void);
void Fog_StopAdditive (void);
void Fog_SetupFrame (void);

extern float turbsin[];
#define TURBSCALE (256.0 / (2 * M_PI))
#define WARPCALC(s,t) ((s + turbsin[(int)((t*2)+(cl.time*(128.0/M_PI))) & 255]) * (1.0/64)) // correct warp

typedef struct
{
	float	x, y, z;
	float	s, t;
	float	r, g, b;
} glvert_t;

extern glvert_t glv;

extern	int glx, gly, glwidth, glheight;

// r_local.h -- private refresh defs

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		1024 //480

#define TILE_SIZE		128		// size of textures generated by tile

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01
#define COLINEAR_EPSILON	0.001

#define FARCLIP		16384 // 4096
#define NEARCLIP		4

extern qboolean r_timerefresh_active;

void R_TimeRefresh_f (void);

void R_ReadPointFile_f (void);

void GL_LightMaps_f (void);
void R_SkyCommand_f (void);
void R_FogCommand_f (void);

texture_t *R_TextureAnimation (texture_t *base, int frame);

typedef struct surfcache_s
{
	struct surfcache_s	*next;
	struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
	int					lightadj[MAXLIGHTMAPS]; // checked for strobe flush
	int					dlight;
	int					size;		// including header
	unsigned			width;
	unsigned			height;		// DEBUG only needed for debug
	float				mipscale;
	struct texture_s	*texture;	// checked for animating textures
	byte				data[4];	// width*height elements
} surfcache_t;

typedef struct
{
	pixel_t		*surfdat;	// destination for generated surface
	int			rowbytes;	// destination logical width in bytes
	msurface_t	*surf;		// description for surface to generate
	fixed8_t	lightadj[MAXLIGHTMAPS];
							// adjust for lightmap levels for dynamic lighting
	texture_t	*texture;	// corrected for animating textures
	int			surfmip;	// mipmapped ratio of surface texels / world pixels
	int			surfwidth;	// in mipmapped texels
	int			surfheight;	// in mipmapped texels
} drawsurf_t;

typedef enum {
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} ptype_t;

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
typedef struct particle_s
{
// driver-usable fields
	vec3_t		org;
	float		color;
// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
} particle_t;

//====================================================

extern	entity_t	r_worldentity;
extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	entity_t	*currententity;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	mplane_t	frustum[4];
extern	int		c_brush_polys, c_alias_polys, c_d_lightmaps;

//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	texture_t	*r_notexture_mip;
extern	texture_t	*r_notexture_mip2;
extern	int		d_lightstylevalue[256];	// 8.8 fraction of base light value

//
//
//
extern int	indexed_bytes;
extern int	lightmap_bytes;
extern int	rgba_bytes;

extern	gltexture_t *notexture;
extern	gltexture_t *nulltexture;

extern	gltexture_t *particletexture;
extern	gltexture_t *particletexture1;
extern	gltexture_t *particletexture2;

extern	gltexture_t	*playertextures[MAX_SCOREBOARD];
extern	gltexture_t	*skyboxtextures[6];

extern	int	gl_warpimagesize;

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_waterquality;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;
extern	cvar_t	r_flatlightstyles;
extern	cvar_t	r_clearcolor;
extern	cvar_t	r_fastsky;
extern	cvar_t	r_skyquality;
extern	cvar_t	r_skyalpha;
extern	cvar_t	r_skyfog;
extern	cvar_t	r_oldsky;

extern	cvar_t	gl_finish;
extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_poly;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_flashblend;

extern	cvar_t  gl_fogenable;
extern	cvar_t  gl_fogdensity;
extern	cvar_t  gl_fogred;
extern	cvar_t  gl_fogblue;
extern	cvar_t  gl_foggreen;

extern	int		gl_lightmap_format;
extern	int		gl_solid_format;
extern	int		gl_alpha_format;

extern	float	r_world_matrix[16];

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;
extern	qboolean fullsbardraw;

// wgl uses APIENTRY
#ifndef APIENTRY
#define APIENTRY
#endif

// for platforms (wgl) that do not use GLAPIENTRY
#ifndef GLAPIENTRY
#define GLAPIENTRY APIENTRY
#endif 

// GL_ARB_multitexture
void (GLAPIENTRY *qglMultiTexCoord1f) (GLenum, GLfloat);
void (GLAPIENTRY *qglMultiTexCoord2f) (GLenum, GLfloat, GLfloat);
void (GLAPIENTRY *qglMultiTexCoord3f) (GLenum, GLfloat, GLfloat, GLfloat);
void (GLAPIENTRY *qglMultiTexCoord4f) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
void (GLAPIENTRY *qglActiveTexture) (GLenum);
void (GLAPIENTRY *qglClientActiveTexture) (GLenum);
#ifndef GL_ACTIVE_TEXTURE_ARB
#define GL_ACTIVE_TEXTURE_ARB			0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE_ARB	0x84E1
#define GL_MAX_TEXTURE_UNITS_ARB		0x84E2
#define GL_TEXTURE0_ARB					0x84C0
#define GL_TEXTURE1_ARB					0x84C1
#define GL_TEXTURE2_ARB					0x84C2
#define GL_TEXTURE3_ARB					0x84C3
#define GL_TEXTURE4_ARB					0x84C4
#define GL_TEXTURE5_ARB					0x84C5
#define GL_TEXTURE6_ARB					0x84C6
#define GL_TEXTURE7_ARB					0x84C7
#define GL_TEXTURE8_ARB					0x84C8
#define GL_TEXTURE9_ARB					0x84C9
#define GL_TEXTURE10_ARB				0x84CA
#define GL_TEXTURE11_ARB				0x84CB
#define GL_TEXTURE12_ARB				0x84CC
#define GL_TEXTURE13_ARB				0x84CD
#define GL_TEXTURE14_ARB				0x84CE
#define GL_TEXTURE15_ARB				0x84CF
#define GL_TEXTURE16_ARB				0x84D0
#define GL_TEXTURE17_ARB				0x84D1
#define GL_TEXTURE18_ARB				0x84D2
#define GL_TEXTURE19_ARB				0x84D3
#define GL_TEXTURE20_ARB				0x84D4
#define GL_TEXTURE21_ARB				0x84D5
#define GL_TEXTURE22_ARB				0x84D6
#define GL_TEXTURE23_ARB				0x84D7
#define GL_TEXTURE24_ARB				0x84D8
#define GL_TEXTURE25_ARB				0x84D9
#define GL_TEXTURE26_ARB				0x84DA
#define GL_TEXTURE27_ARB				0x84DB
#define GL_TEXTURE28_ARB				0x84DC
#define GL_TEXTURE29_ARB				0x84DD
#define GL_TEXTURE30_ARB				0x84DE
#define GL_TEXTURE31_ARB				0x84DF
#endif 

// GL_EXT_texture_env_combine, the values for GL_ARB_ are identical
#define GL_COMBINE_EXT					0x8570
#define GL_COMBINE_RGB_EXT				0x8571
#define GL_COMBINE_ALPHA_EXT			0x8572
#define GL_RGB_SCALE_EXT				0x8573
#define GL_CONSTANT_EXT					0x8576
#define GL_PRIMARY_COLOR_EXT			0x8577
#define GL_PREVIOUS_EXT					0x8578
#define GL_SOURCE0_RGB_EXT				0x8580
#define GL_SOURCE1_RGB_EXT				0x8581
#define GL_SOURCE0_ALPHA_EXT			0x8588
#define GL_SOURCE1_ALPHA_EXT			0x8589

// Multitexture
extern GLenum TEXTURE0, TEXTURE1;

extern qboolean gl_mtexable;
extern qboolean gl_texture_env_combine;
extern qboolean gl_texture_env_add;
extern qboolean mtexenabled;

extern int gl_hardware_maxsize;
extern int gl_max_size;

// Anisotropic Filtering
#define GL_TEXTURE_MAX_ANISOTROPY_EXT		0x84FE
#define GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT	0x84FF

extern float gl_max_anisotropy;
